#include <Arduino.h>
#include <Serial.h>

enum class ErrorCode {
  SUCCESS,
  BUFFER_FULL,
  BUFFER_EMPTY,
  MEMORY_ERROR
};

template<typename T>
class CircularBuffer {
public:
  CircularBuffer(size_t _capacity) : _capacity(_capacity) {
    _buffer = static_cast<T*>(malloc(_capacity * sizeof(T)));
    if (_buffer == nullptr) {
      // Handle memory allocation error
      _errorCode = ErrorCode::MEMORY_ERROR;
    } else {
      _bufferEnd = _buffer + _capacity;
      _count = 0;
      _head = _buffer;
      _tail = _buffer;
      _errorCode = ErrorCode::SUCCESS;
    }
  }

  ~CircularBuffer() {
    free(_buffer);
  }

  ErrorCode pushBack(const T& item) {
    if (_count == _capacity) {
      return ErrorCode::BUFFER_FULL;
    }
    *_head = item;
    _head++;
    if (_head == _bufferEnd) {
      _head = _buffer;
    }
    _count++;
    return ErrorCode::SUCCESS;
  }

  ErrorCode popFront(T& item) {
    if (_count == 0) {
      return ErrorCode::BUFFER_EMPTY;
    }
    item = *_tail;
    _tail++;
    if (_tail == _bufferEnd) {
      _tail = _buffer;
    }
    _count--;
    return ErrorCode::SUCCESS;
  }

private:
  T *_buffer;      // Data _buffer
  T *_bufferEnd;    // End of data buffer
  size_t _capacity;   // Maximum number of items in the buffer
  size_t _count;    // Number of items in the buffer
  T *_head;       // Pointer to head
  T *_tail;       // Pointer to tail
  ErrorCode _errorCode; // Error code
};
